/* Copyright (C) 2000 Massachusetts Institute of Technology.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <stdio.h>

#include "sparsefft3-timers.h"
#include "sparsefft3.h"

#define MIN2(a,b) ((a) < (b) ? (a) : (b))
#define MAX2(a,b) ((a) > (b) ? (a) : (b))
#define MAX3(a,b,c) MAX2(a,MAX2(b,c))

/***************************************************************************/

#ifdef USE_CRAYFFT

#ifdef __cplusplus
extern "C"
#endif                          /* __cplusplus */
extern void CCFFTM(int *isign, int *n, int *lot, double *scale,
                   fftw_complex *x, int *ldx, fftw_complex *y, int *ldy,
                   double *table, double *work, int *isys);

#define PAD_ODD(n) ((n) + (1 - ((n) % 2)))  /* pad n to next odd number */

static void *fftw_malloc(size_t n)
{
     void *p;
     p = malloc(n);
     if (!p) {
	  fprintf(stderr, "sparsefft3_plan.c: out of memory!\n");
	  exit(EXIT_FAILURE);
     }
     return p;
}

#define fftw_free(p) free(p)

static void crayfft_create_plans(crayfft_plan p[3],
				 int n[3], fftw_direction dir)
{
     int i;
     i = MAX3(n[0]*n[1], n[0]*n[2], n[1]*n[2]);
     p[0].work = (double*) fftw_malloc(sizeof(double) * 4 * i);
     p[1].work = p[2].work = p[0].work;
     for (i = 0; i < 3; ++i) {
	  int isign = 0, lot = 1, isys = 0, ldim = 1;
	  double scale = 1.0;
	  fftw_complex dummy = {0,0};
	  p[i].n = n[i];
	  p[i].isign = dir;
	  p[i].table = (double*) fftw_malloc(sizeof(double) * (100+2*n[i]));
	  CCFFTM(&isign, &n[i], &lot, &scale, &dummy, &ldim, &dummy, &ldim,
		 p[i].table, p[i].work, &isys);
	  p[i].free_work = 0;
     }
     p[0].free_work = 1;
}

static void fftw_destroy_plan(fftw_plan p)
{
     fftw_free(p.table);
     if (p.free_work)
	  fftw_free(p.work);
}

#endif

/***************************************************************************/

/* Update r, which points to a linked list of [min,max]
   ranges/intervals, to include the new integers value x (but not any
   other new integers).  Creates a new interval [x,x] if necessary, or
   extends an old one that was adjacent to x.  Upon entry, r may not
   contain any overlapping or adjacent intervals (e.g. [1,2] [3,4]),
   and this property is preserved upon exit. */
static void update_range(sparsefft3_range *r, int x)
{
     sparsefft3_range last_r = NULL; /* the last element of r's linked list */
     sparsefft3_range cur_r = *r;

     while (cur_r) {
	  if (x >= cur_r->min && x <= cur_r->max)
	       return;  /* x is already in an element of r */
	  else if (x == cur_r->min - 1) {
	       cur_r->min--;
	       break;
	  }
	  else if (x == cur_r->max + 1) {
	       cur_r->max++;
	       break;
	  }
	  last_r = cur_r;
	  cur_r = cur_r->next;
     }

     
     if (cur_r) { /* we extended an existing interval */
	  sparsefft3_range cur2_r = cur_r->next, prev_r = cur_r;

	  /* Need to check whether, by extending the existing interval,
	     we have caused this interval to be adjacent with some other
	     interval.  In that case, we can merge the two intervals.
	     Note that we don't have to search intervals prior to
	     cur_r, since we already examined whether x was in them
	     or adjacent to them. */

	  while (cur2_r) {
	       if (cur2_r->min == cur_r->max + 1 || 
		   cur2_r->max == cur_r->min - 1) {
		    /* merge cur2_r into cur_r: */
		    cur_r->max = MAX2(cur_r->max, cur2_r->max);
		    cur_r->min = MIN2(cur_r->min, cur2_r->min);

		    /* Delete cur2_r: */
		    prev_r->next = cur2_r->next;
		    fftw_free(cur2_r);
		    cur2_r = prev_r->next;
	       }
	       else {
		    prev_r = cur2_r;
		    cur2_r = cur2_r->next;
	       }
	  }
     }
     else {  /* need a new interval */
	  sparsefft3_range new_r;

	  new_r =
	       (sparsefft3_range) fftw_malloc(sizeof(sparsefft3_range_struct));
	  new_r->min = new_r->max = x;
	  new_r->next = NULL;
	  if (last_r)
	       last_r->next = new_r;
	  else
	       *r = new_r;
     }     
}

/***************************************************************************/

#ifdef USING_GETTIMEOFDAY

sparsefft3_time sparsefft3_gettimeofday_get_time(void)
{
     struct timeval tv;
     gettimeofday(&tv, 0);
     return tv;
}

sparsefft3_time sparsefft3_gettimeofday_time_diff(sparsefft3_time t1, sparsefft3_time t2)
{
     sparsefft3_time diff;

     diff.tv_sec = t1.tv_sec - t2.tv_sec;
     diff.tv_usec = t1.tv_usec - t2.tv_usec;
     /* normalize */
     while (diff.tv_usec < 0) {
          diff.tv_usec += 1000000L;
          diff.tv_sec -= 1;
     }

     return diff;
}

#endif

double time_sparsefft3(sparsefft3_plan p, fftw_complex *data)
{
     double t;
     TIME_FFT(sparsefft3(p, data), t, SPARSEFFT3_TIME_MIN);
/*     printf("time for %d:%d,%d,%d = %g s\n",
       p->fft2_first, p->first_dim, p->second_dim, p->third_dim, t); */
     return t;
}

/***************************************************************************/

sparsefft3_plan sparsefft3_create_plan_specific(int nx, int ny, int nz,
				       fftw_direction dir, int flags,
				       sparsefft3_sparsedir sparsedir,
				       sparsefft3_nonzero_func nonzero,
				       void *nonzero_data,
				       fftw_complex *data)
{
     sparsefft3_plan p;
     int i, j, ix, iy, iz;

     p = (sparsefft3_plan) fftw_malloc(sizeof(sparsefft3_plan_struct));
     p->sparsedir = sparsedir;
     p->n[0] = nx;
     p->n[1] = ny;
     p->n[2] = nz;
     flags |= FFTW_IN_PLACE;

#ifdef USE_CRAYFFT
     crayfft_create_plans(p->p, p->n, dir);
     p->scratch = (fftw_complex *)
          fftw_malloc(sizeof(fftw_complex) * MAX3(PAD_ODD(nx)*PAD_ODD(ny),
						  PAD_ODD(nx)*PAD_ODD(nz),
						  PAD_ODD(nz)*PAD_ODD(ny)));
#else
     p->scratch = (fftw_complex *)
	  fftw_malloc(sizeof(fftw_complex) * MAX3(nx,ny,nz));
#endif

     for (i = 0; i < 3; ++i) {
#ifndef USE_CRAYFFT
	  p->p[i] = fftw_create_plan(p->n[i], dir, flags);
#endif
	  p->nafter[i] = 1;

	  for (j = 0; j < 3; ++j) {
	       if (j > i)
		    p->nafter[i] *= p->n[j];
	       if (j != i) {
		    int x;
		    p->range1[i][j] = (sparsefft3_range *)
			 fftw_malloc(sizeof(sparsefft3_range) * p->n[i]);
		    for (x = 0; x < p->n[i]; ++x)
			 p->range1[i][j][x] = NULL;
	       }
	       else
		    p->range1[i][j] = NULL;
	  }
	  p->range2[i] = NULL;
     }

     for (ix = 0; ix < nx; ++ix)
	  for (iy = 0; iy < ny; ++iy)
	       for (iz = 0; iz < nz; ++iz) {
		    int x[3];
		    x[0] = ix; x[1] = iy; x[2] = iz;
		    if (nonzero(x, nonzero_data)) {
			 for (i = 0; i < 3; ++i) {
			      update_range(&p->range2[i], x[i]);
			      for (j = 0; j < 3; ++j)
				   if (j != i)
					update_range(&p->range1[i][j][x[i]],
						     x[j]);
			 }
		    }
	       }

     /* Determine the fastest way in which to do the FFT,
	preferably by trying and timing all of the possibilities.
	Or, just guess. */
     if (data != NULL && (flags & FFTW_MEASURE)) {
	  int best_fft2_first = 0, best_first_dim = 0,
	       best_second_dim = 1, best_third_dim = 2;
	  int dim2;
	  double best_time = 1e20;
	  for (dim2 = 0; dim2 < nx*ny*nz; ++dim2)
	       c_re(data[dim2]) = c_im(data[dim2]) = 0.0;
	  for (p->fft2_first = 0; p->fft2_first <= 1; ++p->fft2_first)
	       for (p->first_dim = 0; p->first_dim < 3; ++p->first_dim)
		    for (dim2 = 1; dim2 <= 2; ++dim2) {
			 double t;
			 p->second_dim = (p->first_dim + dim2) % 3;
			 p->third_dim = (p->first_dim + 3-dim2) % 3;
			 t = time_sparsefft3(p, data);
			 if (t < best_time) {
			      best_time = t;
			      best_fft2_first = p->fft2_first;
			      best_first_dim = p->first_dim;
			      best_second_dim = p->second_dim;
			      best_third_dim = p->third_dim;
			 }
		    }
	  p->fft2_first = best_fft2_first;
	  p->first_dim = best_first_dim;
	  p->second_dim = best_second_dim;
	  p->third_dim = best_third_dim;
     }
     else {
#ifdef USE_CRAYFFT
	  p->fft2_first = 1;
	  p->first_dim = 0;
	  p->second_dim = 1;
	  p->third_dim = 2;
#else
	  p->fft2_first = 0;
          p->first_dim = 0;
          p->second_dim = 1;
          p->third_dim = 2;
#endif
     }

     return p;
}

sparsefft3_plan sparsefft3_create_plan(int nx, int ny, int nz,
				       fftw_direction dir, int flags,
				       sparsefft3_sparsedir sparsedir,
				       sparsefft3_nonzero_func nonzero,
				       void *nonzero_data)
{
     return sparsefft3_create_plan_specific(nx, ny, nz, dir, flags, sparsedir,
					    nonzero, nonzero_data, NULL);
}

/***************************************************************************/

static void destroy_range(sparsefft3_range r)
{
     while (r) {
	  sparsefft3_range cur_r = r;
	  r = r->next;
	  free(cur_r);
     }
}

void sparsefft3_destroy_plan(sparsefft3_plan p)
{
     if (p) {
	  int i, j, x;
	  for (i = 0; i < 3; ++i) {
	       fftw_destroy_plan(p->p[i]);
	       destroy_range(p->range2[i]);
	       for (j = 0; j < 3; ++j) {
		    if (p->range1[i][j]) {
			 for (x = 0; x < p->n[i]; ++x)
			      destroy_range(p->range1[i][j][x]);
			 fftw_free(p->range1[i][j]);
		    }
	       }
	  }
	  fftw_free(p->scratch);
	  fftw_free(p);
     }
}

/***************************************************************************/
