/* Copyright (C) 2000 Massachusetts Institute of Technology.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "sparsefft3.h"
#include "sparsefft3-timers.h"

typedef struct {
     double u0, u1, u2, r;
     int n0,n1,n2;
} insphere_data;

int insphere(int x[3], void *data)
{
     insphere_data *d = (insphere_data *) data;
     double x2, y2, z2, mx2, my2, mz2, r2;
     r2 = d->r * d->r;
     x2 = d->u0 * d->u0 * x[0]*x[0];
     y2 = d->u1 * d->u1 * x[1]*x[1];
     z2 = d->u2 * d->u2 * x[2]*x[2];
     mx2 = d->u0 * d->u0 * (d->n0 - x[0]) * (d->n0 - x[0]);
     my2 = d->u1 * d->u1 * (d->n1 - x[1]) * (d->n1 - x[1]);
     mz2 = d->u2 * d->u2 * (d->n2 - x[2]) * (d->n2 - x[2]);
     return(x2 + y2 + z2 <= r2 ||
	    x2 + y2 + mz2 <= r2 ||
	    x2 + my2 + z2 <= r2 ||
	    x2 + my2 + mz2 <= r2 ||
	    mx2 + y2 + z2 <= r2 ||
	    mx2 + y2 + mz2 <= r2 ||
	    mx2 + my2 + z2 <= r2 ||
	    mx2 + my2 + mz2 <= r2);
}

int use_inrandom = 0;

int inrandom(int x[3], void *data)
{
     double *prob = (double *) data;
     srand(x[0] * 123427 + x[1] * 293173 + x[2] * 387173);
     return rand() * 1.0 / RAND_MAX <= *prob;
}

void display_range(sparsefft3_range r)
{
     if (!r)
	  printf("  []\n");
     else {
	  while (r) {
	       printf("  [%d,%d]", r->min, r->max);
	       r = r->next;
	  }
	  printf("\n");
     }
}

#ifdef USE_CRAYFFT
extern void CCFFT3D(int *isign, int *n1, int *n2, int *n3, double *scale,
		    fftw_complex *x, int *ldx, int *ldx2,
		    fftw_complex *y, int *ldy, int *ldy2,
		    double *table, double *work, int *isys);
/* table: 100 + 2*(n1+n2+n3).
   work: 512 * MAX(n1,n2,n3) when isys = 0 or
         4 * NCPUS * MAX(n1*n2, n2*n3, n3*n1) when isys = 1 (faster)
   NCPUS = atoi(getenv("NCPUS") ? getenv("NCPUS") : "4") */
#endif

#ifdef FFTW_ENABLE_FLOAT
#  define TOLERANCE (1e-2)
#else
#  define TOLERANCE (1e-6)
#endif

/***********************************************************************/

/* return a randum real number in [xmin,xmax] */
double rand_x(double xmin, double xmax)
{
     double d = rand();
     return (xmin + (xmax - xmin) * (d / (double) RAND_MAX));
}

/* return a random integer in [xmin, xmax] */
int rand_n(int xmin, int xmax)
{
     return (rand() % (xmax - xmin + 1) + xmin);
}

#define SQR(x) ((x) * (x))

double compute_error_complex(fftw_complex * A, fftw_complex * B, int n)
{
     /* compute the relative error */
     double error = 0.0;
     int i;

     for (i = 0; i < n; ++i) {
          double a;
          double mag;
          a = sqrt(SQR(c_re(A[i]) - c_re(B[i])) +
                   SQR(c_im(A[i]) - c_im(B[i])));
          mag = 0.5 * (sqrt(SQR(c_re(A[i])) + SQR(c_im(A[i]))) +
                       sqrt(SQR(c_re(B[i])) + SQR(c_im(B[i])))) + TOLERANCE;

          a /= mag;
          if (a > error)
               error = a;

#ifdef HAVE_ISNAN
          CHECK(!isnan(a), "NaN in answer");
#endif
     }
     return error;
}

void zero_arr(int n, fftw_complex *a)
{
     int i;
     for (i = 0; i < n; ++i)
	  c_re(a[i]) = c_im(a[i]) = 0.0;
}

#define MAX2(a,b) ((a) > (b) ? (a) : (b))
#define MAX3(a,b,c) MAX2(a,MAX2(b,c))

void sparsefft3_randomize_plan(sparsefft3_plan p)
{
     int dim2;
     p->fft2_first = rand_n(0,1);
     p->first_dim = rand_n(0,2);
     dim2 = rand_n(1,2);
     p->second_dim = (p->first_dim + dim2) % 3;
     p->third_dim = (p->first_dim + 3 - dim2) % 3;
}

double sparsefft3_test(int nx, int ny, int nz,
		    fftw_direction dir,
		    double ux, double uy, double uz,
		    double r, int time_fft, int specific)
{
     sparsefft3_plan sp, isp;
#ifdef USE_CRAYFFT
     int isign, isys;
     double *table, *work, scale = 1.0;
#else
     fftwnd_plan p;
#endif
     fftw_complex *a, *b;
     int ix, iy, iz, i, N = nx*ny*nz;
     double ux2 = ux*ux/(nx*nx), uy2 = uy*uy/(ny*ny), uz2 = uz*uz/(nz*nz);
     double r2 = r*r;
     double err;
     double t = 1, st = 1;
     int fill_N = 0;
     insphere_data isd;
     double prob = 0.3;
     int flags;
     sparsefft3_nonzero_func func;
     void *func_data;

     isd.n0 = nx; isd.n1 = ny; isd.n2 = nz;
     isd.u0 = ux/nx; isd.u1 = uy/ny; isd.u2 = uz/nz;
     isd.r = r;

     if (!use_inrandom) {
	  func = insphere;
	  func_data = &isd;
     }
     else {
	  func = inrandom;
	  func_data = &prob;
     }

     if (specific)
	  printf("s: ");
     printf("%dx%dx%d, (%g,%g,%g), r=%g...",
	    nx,ny,nz, ux,uy,uz, r);
     fflush(stdout);

     a = (fftw_complex*) malloc(sizeof(fftw_complex) * N);
     b = (fftw_complex*) malloc(sizeof(fftw_complex) * N);

     if (time_fft)
	  flags = FFTW_MEASURE;
     else
	  flags = FFTW_ESTIMATE;
     flags |= FFTW_USE_WISDOM;

     sp = sparsefft3_create_plan_specific(nx, ny, nz, dir, flags,
					  SPARSEFFT3_SPARSEINPUT,
					  func,func_data, specific ? a : NULL);
     isp = sparsefft3_create_plan_specific(nx, ny, nz, dir, flags,
					  SPARSEFFT3_SPARSEOUTPUT,
					  func,func_data, specific ? a : NULL);

     /* If we're not benchmarking, randomize the plan so that
	we test all the different variations: */
     if (!time_fft) {
	  sparsefft3_randomize_plan(sp);
	  sparsefft3_randomize_plan(isp);
     }

#if USE_CRAYFFT
     isign = 0;
     table = (double*) malloc(sizeof(double) * (100 + 2*(nx+ny+nz)));
     work = (double*) malloc(sizeof(double) * 512 * MAX3(nx,ny,nz));
     isys = 0;
     /* initialize trig tables: */
     CCFFT3D(&isign, &nz, &ny, &nx, &scale,
	     b, &nz, &ny, b, &nz, &ny, table, work, &isys);
     isign = (int) dir;
#else
     p = fftw3d_create_plan(nx, ny, nz, dir, 
			    flags | FFTW_IN_PLACE);
#endif

     for (ix = 0; ix < nx; ++ix)
	  for (iy = 0; iy < ny; ++iy)
	       for (iz = 0; iz < nz; ++iz) {
		    int x[3];
		    x[0] = ix; x[1] = iy; x[2] = iz;
		    if (insphere(x, &isd)) {
			 c_re(a[(ix*ny + iy)*nz + iz]) = rand_x(-1,1);
			 c_im(a[(ix*ny + iy)*nz + iz]) = rand_x(-1,1);
			 ++fill_N;
		    }
		    else 
			 c_re(a[(ix*ny + iy)*nz + iz]) =
			      c_im(a[(ix*ny + iy)*nz + iz]) = 0.0;
		    b[(ix*ny + iy)*nz + iz] = a[(ix*ny + iy)*nz + iz];
	       }

     printf("%.2f%% fill...", fill_N * 100.0 / N);
     fflush(stdout);

#ifdef USE_CRAYFFT
     CCFFT3D(&isign, &nz, &ny, &nx, &scale,
	     b, &nz, &ny, b, &nz, &ny, table, work, &isys);
#else
     fftwnd_one(p, b, b);
#endif
     sparsefft3(sp, a);

     if ((err = compute_error_complex(a, b, N)) > TOLERANCE) {
	  fprintf(stderr, "Large error %g in sparsefft3!\n", err);
	  exit(EXIT_FAILURE);
     }

     printf("ok...");
     fflush(stdout);

     for (i = 0; i < N; ++i) {
	  c_re(a[i]) = rand_x(-1,1);
	  c_im(a[i]) = rand_x(-1,1);
	  b[i] = a[i];
     }

#ifdef USE_CRAYFFT
     CCFFT3D(&isign, &nz, &ny, &nx, &scale,
	     b, &nz, &ny, b, &nz, &ny, table, work, &isys);
#else
     fftwnd_one(p, b, b);
#endif
     sparsefft3(isp, a);

     /* Compute maximum error of output inside the sphere: */
     err = 0;
     for (ix = 0; ix < nx; ++ix)
	  for (iy = 0; iy < ny; ++iy)
	       for (iz = 0; iz < nz; ++iz) {
		    int x[3];
		    x[0] = ix; x[1] = iy; x[2] = iz;
		    if (insphere(x, &isd)) {
			 double err2;
			 err2 = compute_error_complex(&a[(ix*ny + iy)*nz + iz],
						      &b[(ix*ny + iy)*nz + iz],
						      1);
			 err = MAX2(err2,err);
		    }
	       }
     if (err > TOLERANCE) {
	  fprintf(stderr, "Large error %g in isparsefft3!\n", err);
	  exit(EXIT_FAILURE);
     }

     printf("ok.\n");
     fflush(stdout);

     zero_arr(N, a);
     zero_arr(N, b);
     if (time_fft) {
#ifdef USE_CRAYFFT
	  TIME_FFT(CCFFT3D(&isign, &nz, &ny, &nx, &scale,
			   b, &nz, &ny, b, &nz, &ny, table, work, &isys),
		   t,SPARSEFFT3_TIME_MIN*20);
#else
	  TIME_FFT(fftwnd_one(p, b, b), t, SPARSEFFT3_TIME_MIN*20);
#endif

	  TIME_FFT(sparsefft3(sp, a), st, SPARSEFFT3_TIME_MIN*20);
     }

     if (time_fft) {
	  if (specific)
	       printf("s: ");
	  printf("fft2_first=%d, dim ordering = %d, %d, %d\n",
		 sp->fft2_first, sp->first_dim, sp->second_dim,
		 sp->third_dim);
	  if (specific)
	       printf("s: ");
	  printf("sphere/normal: "
		 "%g/%g s, %g/%g mflops, %g speedup\n",
		 st, t,
		 5*N*log(N)/log(2) / (1e6*st), 5*N*log(N)/log(2) / (1e6*t),
		 t / st);
     }

#ifdef USE_CRAYFFT
     free(table);
     free(work);
#else
     fftwnd_destroy_plan(p);
#endif

     sparsefft3_destroy_plan(sp);
     sparsefft3_destroy_plan(isp);

     free(b);
     free(a);

     return (t/st);
}

extern void fftw_check_memory_leaks(void);

#define SIZETIME 128
#define NTEST 300

void usage(FILE *f)
{
     fprintf(f,
	     "usage: sparsefft3_test [options]\n"
	     "Options:\n"
	     "    -h : this help\n"
	     "-c <n> : test correctness for n times (default %d)\n"
	     "-s <n> : test speed for n x n x n (default %d)\n",
	     NTEST, SIZETIME);
}

int main(int argc, char **argv)
{
     int i, ntest = NTEST, sizetime = SIZETIME;
     int testspeed = 0, testcorrect = 0;

     srand(time(NULL));

     for (i = 1; i < argc; ++i) {
	  if (!strcmp(argv[i], "-h")) {
	       usage(stdout);
	       return EXIT_SUCCESS;
	  }
	  else if (!strcmp(argv[i], "-r")) {
	       use_inrandom = 1;
	  }
	  else if (!strcmp(argv[i], "-s")) {
	       testspeed = 1;
	       if (i + 1 < argc && argv[i + 1][0] != '-')
		    sizetime = atoi(argv[++i]);
	  }
	  else if (!strcmp(argv[i], "-S")) {
	       testspeed = 10;
	       if (i + 1 < argc && argv[i + 1][0] != '-')
		    sizetime = atoi(argv[++i]);
	  }
	  else if (!strcmp(argv[i], "-c")) {
	       testcorrect = 1;
	       if (i + 1 < argc && argv[i + 1][0] != '-')
		    ntest = atoi(argv[++i]);
	  }
	  else {
	       usage(stderr);
	       return EXIT_FAILURE;
	  }
     }

     if (argc < 2) {
	  usage(stderr);
	  return EXIT_FAILURE;
     }

     if (testspeed && sizetime > 0) {
	  double speedup_sum = 0, speedup_sqr_sum = 0;
	  double speedup2_sum = 0, speedup2_sqr_sum = 0;
	  for (i = 0; i < testspeed; ++i) {
	       double t;
	       t = sparsefft3_test(sizetime,sizetime,sizetime,
			       FFTW_FORWARD, 1,1,1, .25, 1, 0);
	       speedup_sum += t;
	       speedup_sqr_sum += t*t;
	       t = sparsefft3_test(sizetime,sizetime,sizetime,
				   FFTW_FORWARD, 1,1,1, .25, 1, 1);
	       speedup2_sum += t;
	       speedup2_sqr_sum += t*t;
	  }
	  printf("  normal: mean speedup = %g +- %g\n",
		 speedup_sum/testspeed,
		 sqrt(speedup_sqr_sum/testspeed -
		      (speedup_sum/testspeed) * (speedup_sum/testspeed)));
	  printf("specific: mean speedup = %g +- %g\n",
		 speedup2_sum/testspeed,
		 sqrt(speedup2_sqr_sum/testspeed -
		      (speedup2_sum/testspeed) * (speedup2_sum/testspeed)));
     }

     if (testcorrect)
	  for (i = 0; i < ntest; ++i) {
	       sparsefft3_test(rand_n(20,100), rand_n(20,100), rand_n(20,100),
			      rand_n(0,1) ? FFTW_FORWARD : FFTW_BACKWARD,
			      rand_x(0.5,1.5),rand_x(0.5,1.5),rand_x(0.5,1.5),
			      rand_x(0.1,0.6), 0, 0);
	  }
     
#ifndef USE_CRAYFFT
     fftw_check_memory_leaks();
#endif
     
     printf("Finished all tests successfully.\n");

     return EXIT_SUCCESS;
}
